# coding=utf-8

import logging

from django.test import TestCase
from django.utils import timezone

from web.importer import Importer
from web.models import Subject, StudySubject, Provenance, SubjectImportData
from web.tests.functions import create_study_subject, get_test_study
from .mock_reader import MockReader

logger = logging.getLogger(__name__)


class TestImporter(TestCase):

    def setUp(self):
        self.study = get_test_study()
        self.visit_import_data = SubjectImportData.objects.create(study=self.study)

    def test_import_new_subject(self):
        study_subjects = []
        subject = Subject()
        subject.first_name = "Piotr"
        subject.last_name = "Gawron"
        study_subject = StudySubject()
        study_subject.screening_number = "Cov-123456"
        study_subject.subject = subject
        study_subject.study = self.study
        study_subjects.append(study_subject)

        importer = Importer(reader=MockReader(self.visit_import_data, study_subjects))
        subject_counter = Subject.objects.count()
        study_subject_counter = StudySubject.objects.count()
        provenance_counter = Provenance.objects.count()
        importer.execute()

        self.assertEqual(subject_counter + 1, Subject.objects.count())
        self.assertEqual(study_subject_counter + 1, StudySubject.objects.count())
        self.assertNotEqual(provenance_counter, Provenance.objects.count())

        self.assertEqual(1, importer.added_count)
        self.assertEqual(0, importer.problematic_count)
        self.assertEqual(0, importer.warning_count)

    def test_import_invalid(self):
        study_subjects = []
        study_subject = StudySubject()
        study_subject.screening_number = "Cov-123456"
        study_subject.study = self.study
        study_subjects.append(study_subject)

        importer = Importer(reader=MockReader(self.visit_import_data, study_subjects))
        study_subject_counter = StudySubject.objects.count()
        importer.execute()

        self.assertEqual(study_subject_counter, StudySubject.objects.count())

        self.assertEqual(0, importer.added_count)
        self.assertEqual(1, importer.problematic_count)

    def test_import_no_study(self):
        study_subjects = []
        study_subject = StudySubject()
        study_subject.screening_number = "Cov-123456"
        study_subjects.append(study_subject)

        importer = Importer(reader=MockReader(self.visit_import_data, study_subjects))
        study_subject_counter = StudySubject.objects.count()
        importer.execute()

        self.assertEqual(study_subject_counter, StudySubject.objects.count())

        self.assertEqual(0, importer.added_count)
        self.assertEqual(1, importer.problematic_count)

    def test_import_merge_subject(self):
        existing_study_subject = create_study_subject()
        study_subjects = []
        subject = Subject()
        subject.first_name = "XYZ"
        subject.last_name = "AAA"
        subject.date_born = timezone.now()
        study_subject = StudySubject()
        study_subject.screening_number = existing_study_subject.screening_number
        study_subject.nd_number = existing_study_subject.nd_number
        study_subject.subject = subject
        study_subject.study = self.study
        study_subjects.append(study_subject)

        importer = Importer(reader=MockReader(self.visit_import_data, study_subjects))
        provenance_counter = Provenance.objects.count()
        subject_counter = Subject.objects.count()
        study_subject_counter = StudySubject.objects.count()
        importer.execute()

        self.assertEqual(subject_counter, Subject.objects.count())
        self.assertEqual(study_subject_counter, StudySubject.objects.count())
        self.assertNotEqual(provenance_counter, Provenance.objects.count())

        self.assertEqual(0, importer.added_count)
        self.assertEqual(1, importer.merged_count)
        self.assertEqual(0, importer.problematic_count)
        self.assertEqual(0, importer.warning_count)

        existing_study_subject = StudySubject.objects.filter(id=existing_study_subject.id)[0]
        self.assertEqual(existing_study_subject.subject.first_name, subject.first_name)
        self.assertEqual(existing_study_subject.subject.last_name, subject.last_name)
        self.assertEqual(existing_study_subject.subject.date_born.strftime("%Y-%m-%d"),
                         subject.date_born.strftime("%Y-%m-%d"))
