# coding=utf-8
from django.utils import timezone
import datetime
from datetime import timedelta

from web.algorithm import VerhoeffAlgorithm, LuhnAlgorithm

from django.db import models
from django.contrib.admin.utils import NestedObjects
from django.utils.text import capfirst
from django.utils.encoding import force_text
from typing import List

def get_deleted_objects(objs: List[models.Model]): 
    collector = NestedObjects(using='default')
    collector.collect(objs)
    #
    def format_callback(obj):
        opts = obj._meta
        no_edit_link = '%s: %s' % (capfirst(opts.verbose_name),
                                   force_text(obj))
        return no_edit_link            
    #
    to_delete = collector.nested(format_callback)
    protected = [format_callback(obj) for obj in collector.protected]
    model_count = {model._meta.verbose_name_plural: len(objs) for model, objs in collector.model_objs.items()}
    #
    return to_delete, model_count, protected

def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip

def is_valid_social_security_number(number):
    if number is not None and number != '':
        if len(number) != 13:
            return False
        if not number.isdigit():
            return False
        if not LuhnAlgorithm.is_luhn_valid(number[:12]):
            return False
        if not VerhoeffAlgorithm.is_valid_verhoeff(number[:11] + number[12]):
            return False

    return True


def get_today_midnight_date():
    today = timezone.now()
    today_midnight = datetime.datetime(
        today.year, 
        today.month, 
        today.day, 
        tzinfo=today.tzinfo
    )
    return today_midnight


def get_weekdays_in_period(fromdate, todate):
    '''
    fromdate and todate must be generated using datetime.date or datetime.datetime like:

    from datetime import date
    fromdate = date(2010,1,1)
    todate = date(2010,3,31)

    fromdate = datetime.datetime(2018, 10, 3, 15, 00, 00)
    todate = datetime.datetime.today()

    but both dates must have the same format ! 
    todate is not included in the range

    Weekdays are returned as isoweekdays like the form described in week_choices from constants.py (starting at 1)
    '''
    if todate < fromdate:
        return set([])
    day_generator = (fromdate + timedelta(day)
                     for day in range((todate - fromdate).days))
    weekdays = set([date.isoweekday() for date in day_generator])
    return weekdays
